from typing import Any, Iterable, Tuple, Type, Union

from boltons.urlutils import URL
from ereuse_utils.test import Client as EreuseUtilsClient, JSON, Res
from werkzeug.exceptions import HTTPException

from teal.marshmallow import ValidationError

Status = Union[int, Type[HTTPException], Type[ValidationError]]
Query = Iterable[Tuple[str, Any]]


class Client(EreuseUtilsClient):
    """A REST interface to a Teal app."""

    def open(self,
             uri: str,
             res: str = None,
             status: Status = 200,
             query: Query = tuple(),
             accept=JSON,
             content_type=JSON,
             item=None,
             headers: dict = None,
             token: str = None, **kw) -> Res:
        headers = headers or {}
        if res:
            resource_url = self.application.resources[res].url_prefix + '/'
            uri = URL(uri).navigate(resource_url).to_text()
        if token:
            headers['Authorization'] = 'Basic {}'.format(token)
        res = super().open(uri, status, query, accept, content_type, item, headers, **kw)
        # ereuse-utils checks for status code
        # here we check for specific type
        # (when response: {'type': 'foobar', 'code': 422})
        _status = getattr(status, 'code', status)
        if not isinstance(status, int) and res[1].status_code == _status:
            assert status.__name__ == res[0]['type'], \
                'Expected exception {0} but it was {1}'.format(status.__name__, res[0]['type'])
        return res

    def get(self,
            uri: str = '',
            res: str = None,
            query: Query = tuple(),
            status: Status = 200,
            item=None,
            accept: str = JSON,
            headers: dict = None,
            token: str = None,
            **kw) -> Res:
        """
        Performs GET.

        :param uri: The uri where to GET from. This is optional, as you
                    can build the URI too through ``res`` and ``item``.
        :param res: The resource where to GET from, if any.
                    If this is set, the client will try to get the
                    url from the resource definition.
        :param query: The query params in a dict. This method
                      automatically converts the dict to URL params,
                      and if the dict had nested dictionaries, those
                      are converted to JSON.
        :param status: A status code or exception to assert.
        :param item: The id of a resource to GET from, if any.
        :param accept: The accept headers. By default
                       ``application/json``.
        :param headers: A dictionary of header name - header value.
        :param token: A token to add to an ``Authentication`` header.
        :return: A tuple containing 1. a dict (if content-type is JSON)
                 or a str with the data, and 2. the ``Response`` object.
        """
        kw['res'] = res
        kw['token'] = token
        return super().get(uri, query, item, status, accept, headers, **kw)

    def post(self,
             data: str or dict,
             uri: str = '',
             res: str = None,
             query: Query = tuple(),
             status: Status = 201,
             content_type: str = JSON,
             accept: str = JSON,
             headers: dict = None,
             token: str = None,
             **kw) -> Res:
        kw['res'] = res
        kw['token'] = token
        return super().post(uri, data, query, status, content_type, accept, headers, **kw)

    def patch(self,
              data: str or dict,
              uri: str = '',
              res: str = None,
              query: Query = tuple(),
              item=None,
              status: Status = 200,
              content_type: str = JSON,
              accept: str = JSON,
              token: str = None,
              headers: dict = None,
              **kw) -> Res:
        kw['res'] = res
        kw['token'] = token
        return super().patch(uri, data, query, status, content_type, item, accept, headers, **kw)

    def put(self,
            data: str or dict,
            uri: str = '',
            res: str = None,
            query: Query = tuple(),
            item=None,
            status: Status = 201,
            content_type: str = JSON,
            accept: str = JSON,
            token: str = None,
            headers: dict = None,
            **kw) -> Res:
        kw['res'] = res
        kw['token'] = token
        return super().put(uri, data, query, status, content_type, item, accept, headers, **kw)

    def delete(self,
               uri: str = '',
               res: str = None,
               query: Query = tuple(),
               status: Status = 204,
               item=None,
               accept: str = JSON,
               headers: dict = None,
               token: str = None,
               **kw) -> Res:
        kw['res'] = res
        kw['token'] = token
        return super().delete(uri, query, item, status, accept, headers, **kw)

    def post_get(self,
                 res: str,
                 data: str or dict,
                 query: Query = tuple(),
                 status: Status = 200,
                 content_type: str = JSON,
                 accept: str = JSON,
                 headers: dict = None,
                 key='id',
                 token: str = None, **kw) -> Res:
        """Performs post and then gets the resource through its key."""
        r, _ = self.post('', data, res, query, status, content_type, accept, token, headers, **kw)
        return self.get(res=res, item=r[key])



import inspect
from typing import Dict, Iterable, Tuple, Type

import click_spinner
import flask_cors
from anytree import Node
from apispec import APISpec
from click import option
from ereuse_utils import ensure_utf8
from flask import Flask, Response, jsonify
from flask.globals import _app_ctx_stack
from flask_sqlalchemy import SQLAlchemy
from marshmallow import ValidationError
from werkzeug.exceptions import HTTPException, UnprocessableEntity
from werkzeug.wsgi import DispatcherMiddleware

from teal.auth import Auth
from teal.client import Client
from teal.config import Config as ConfigClass
from teal.json_util import TealJSONEncoder
from teal.request import Request
from teal.resource import Resource


class Teal(Flask):
    """
    An opinionated REST and JSON first server built on Flask using
    MongoDB and Marshmallow.
    """
    test_client_class = Client
    request_class = Request
    json_encoder = TealJSONEncoder

    def __init__(self,
                 config: ConfigClass,
                 db: SQLAlchemy,
                 import_name=__name__.split('.')[0],
                 static_url_path=None,
                 static_folder='static',
                 static_host=None,
                 host_matching=False,
                 subdomain_matching=False,
                 template_folder='templates',
                 instance_path=None,
                 instance_relative_config=False,
                 root_path=None,
                 Auth: Type[Auth] = Auth):
        ensure_utf8(self.__class__.__name__)
        super().__init__(import_name, static_url_path, static_folder, static_host, host_matching,
                         subdomain_matching, template_folder, instance_path,
                         instance_relative_config, root_path)
        self.config.from_object(config)
        flask_cors.CORS(self)
        # Load databases
        self.auth = Auth()
        self.load_resources()
        self.register_error_handler(HTTPException, self._handle_standard_error)
        self.register_error_handler(ValidationError, self._handle_validation_error)
        self.db = db
        db.init_app(self)
        self.cli.command('init-db')(self.init_db)
        self.spec = None  # type: APISpec
        self.apidocs()

    # noinspection PyAttributeOutsideInit
    def load_resources(self):
        self.resources = {}  # type: Dict[str, Resource]
        """
        The resources definitions loaded on this App, referenced by their
        type name.
        """
        self.tree = {}  # type: Dict[str, Node]
        """
        A tree representing the hierarchy of the instances of 
        ResourceDefinitions. ResourceDefinitions use these nodes to
        traverse their hierarchy.
         
        Do not use the normal python class hierarchy as it is global,
        thus unreliable if you run different apps with different
        schemas (for example, an extension that is only added on the
        third app adds a new type of user).
        """
        for ResourceDef in self.config['RESOURCE_DEFINITIONS']:
            resource_def = ResourceDef(self)  # type: Resource
            self.register_blueprint(resource_def)
            for cli_command, *args in resource_def.cli_commands:  # Register CLI commands
                self.cli.command(*args)(cli_command)

            # todo should we use resource_def.name instead of type?
            # are we going to have collisions? (2 resource_def -> 1 schema)
            self.resources[resource_def.type] = resource_def
            self.tree[resource_def.type] = Node(resource_def.type)
        # Link tree nodes between them
        for _type, node in self.tree.items():
            resource_def = self.resources[_type]
            _, Parent, *superclasses = inspect.getmro(resource_def.__class__)
            if Parent is not Resource:
                node.parent = self.tree[Parent.type]

    @staticmethod
    def _handle_standard_error(e: HTTPException):
        """
        Handles HTTPExceptions by transforming them to JSON.
        """
        try:
            data = {'message': e.description, 'code': e.code, 'type': e.__class__.__name__}
        except AttributeError as e:
            return Response(str(e), status=500)
        else:
            response = jsonify(data)
            response.status_code = e.code
            return response

    @staticmethod
    def _handle_validation_error(e: ValidationError):
        data = {
            'message': e.messages,
            'code': UnprocessableEntity.code,
            'type': e.__class__.__name__
        }
        response = jsonify(data)
        response.status_code = UnprocessableEntity.code
        return response

    @option('--erase/--no-erase', default=False, help='Delete all db contents before?')
    def init_db(self, erase: bool = False):
        """
        Initializes a database from scratch,
        creating tables and needed resources.

        Note that this does not create the database per se.

        If executing this directly, remember to use an app_context.

        Resources can hook functions that will be called when this
        method executes, by subclassing :meth:`teal.resource.
        Resource.load_resource`.
        """
        assert _app_ctx_stack.top, 'Use an app context.'
        print('Initializing database...'.ljust(30), end='')
        with click_spinner.spinner():
            if erase:
                self.db.drop_all()
            self._init_db()
            self.db.session.commit()

    def _init_db(self):
        """Where the database is initialized. You can override this."""
        self.db.create_all()
        for resource in self.resources.values():
            resource.init_db(self.db)

    def apidocs(self):
        """Apidocs configuration and generation."""
        self.spec = APISpec(
            plugins=(
                'apispec.ext.flask',
                'apispec.ext.marshmallow',
            ),
            **self.config.get_namespace('API_DOC_CONFIG_')
        )
        for name, resource in self.resources.items():
            self.spec.definition(name,
                                 schema=resource.SCHEMA,
                                 extra_fields=self.config.get_namespace('API_DOC_CLASS_'))
        self.add_url_rule('/apidocs', view_func=self.apidocs_endpoint)

    def apidocs_endpoint(self):
        """An endpoint that prints a JSON OpenApi 2.0 specification."""
        if not getattr(self, '_apidocs', None):
            # We are forced to to this under a request context
            for path, view_func in self.view_functions.items():
                if path != 'static':
                    self.spec.add_path(view=view_func)
            self._apidocs = self.spec.to_dict()
        return jsonify(self._apidocs)


def prefixed_database_factory(Config: Type[ConfigClass],
                              databases: Iterable[Tuple[str, str]],
                              App: Type[Teal] = Teal) -> DispatcherMiddleware:
    """
    A factory of Teals. Allows creating as many Teal apps as databases
    from the DefaultConfig.DATABASES, setting each Teal app to an URL in
    the following way:
    - / -> to the Teal app that uses the
      :attr:`teal.config.Config.SQLALCHEMY_DATABASE_URI` set in config.
    - /db1/... -> to the Teal app with db1 as default
    - /db2/... -> to the Teal app with db2 as default
    And so on.

    DefaultConfig is used to configure the root Teal app.
    Optionally, each other app can have a custom Config. Pass them in
    the `configs` dictionary. Apps with no Config will default to the
    DefaultConfig.

    :param Config: The configuration class to use with each database
    :param databases: Names of the databases, where the first value is a
                      valid  URI to use in the dispatcher middleware and
                      the second value the SQLAlchemy URI referring to a
                      database to use.
    :param App: A Teal class.
    :return: A WSGI middleware where an app without database is default
    and the rest prefixed with their database name.
    """
    # todo
    db = SQLAlchemy()
    default = App(config=Config(), db=db)
    apps = {
        '/{}'.format(path_uri): App(config=Config(db=sql_uri), db=db)
        for path_uri, sql_uri in databases
    }
    return DispatcherMiddleware(default, apps)



from contextlib import contextmanager

import pytest
from flask.testing import FlaskCliRunner
from marshmallow.fields import Integer, Str
from marshmallow.validate import Range

from teal.client import Client
from teal.config import Config
from teal.db import INHERIT_COND, Model, POLYMORPHIC_ID, POLYMORPHIC_ON, SQLAlchemy
from teal.marshmallow import NestedOn
from teal.resource import Converters, Resource, Schema, View
from teal.teal import Teal


@pytest.fixture()
def config() -> Config:
    class TestConfig(Config):
        Testing = True
        SQLALCHEMY_DATABASE_URI = 'sqlite:///:memory:'
        SQLALCHEMY_TRACK_MODIFICATIONS = False

    return TestConfig()


@pytest.fixture()
def db():
    return SQLAlchemy(model_class=Model)


@pytest.fixture()
def fconfig(config: Config, db: SQLAlchemy) -> Config:
    return f_config(config, db)


def f_config(config: Config, db: SQLAlchemy) -> Config:
    """
    Creates 3 resources like in the following::

        Device
           |
           |________
          /        |
        Computer Component

    ``Computer`` and ``Component`` inherit from ``Device``, and
    a computer has many components::

        Computer 1 -- * Component
    """

    class DeviceSchema(Schema):
        id = Integer(validate=Range(min=1))
        type = Str()
        model = Str()

    class DeviceView(View):
        pass

    class Device(db.Model):
        id = db.Column(db.Integer, primary_key=True)
        model = db.Column(db.String(80), nullable=True)
        type = db.Column(db.String)

        __mapper_args__ = {
            POLYMORPHIC_ID: 'Device',
            POLYMORPHIC_ON: type
        }

    class DeviceDef(Resource):
        SCHEMA = DeviceSchema
        VIEW = DeviceView
        MODEL = Device
        ID_CONVERTER = Converters.int

    class ComponentSchema(DeviceSchema):
        pass

    class Component(Device):
        id = db.Column(db.Integer, db.ForeignKey(Device.id), primary_key=True)

        parent_id = db.Column(db.Integer, db.ForeignKey('computer.id'))
        parent = db.relationship('Computer',
                                 backref=db.backref('components', lazy=True),
                                 primaryjoin='Component.parent_id == Computer.id')

        __mapper_args__ = {
            POLYMORPHIC_ID: 'Component',
            INHERIT_COND: id == Device.id
        }

    class ComponentView(DeviceView):
        pass

    class ComponentDef(DeviceDef):
        SCHEMA = ComponentSchema
        VIEW = ComponentView
        MODEL = Component

    class ComputerSchema(DeviceSchema):
        components = NestedOn(ComponentSchema, polymorphic_on='type', many=True, db=db)

    class Computer(Device):
        id = db.Column(db.Integer, db.ForeignKey(Device.id), primary_key=True)
        # backref creates a 'parent' relationship in Component

        __mapper_args__ = {
            POLYMORPHIC_ID: 'Computer',
            INHERIT_COND: id == Device.id
        }

    class ComputerView(DeviceView):
        pass

    class ComputerDef(DeviceDef):
        SCHEMA = ComputerSchema
        VIEW = ComputerView
        MODEL = Computer

    config.RESOURCE_DEFINITIONS = DeviceDef, ComponentDef, ComputerDef
    return config


@pytest.fixture()
def app(fconfig: Config, db: SQLAlchemy) -> Teal:
    app = Teal(config=fconfig, db=db)
    with app.app_context():
        app.init_db()
    yield app


@pytest.fixture()
def app_context(app: Teal):
    with app.app_context():
        yield


@pytest.fixture()
def client(app: Teal) -> Client:
    return app.test_client()


@contextmanager
def populated_db(db: SQLAlchemy, app: Teal):
    db.create_all(app=app)
    yield


@pytest.fixture()
def runner(app: Teal) -> FlaskCliRunner:
    return app.test_cli_runner()



import ipaddress
from distutils import version

import colour
import pytest
from boltons import urlutils
from marshmallow import Schema
from sqlalchemy_utils import PhoneNumber, PhoneNumberParseException

from teal import marshmallow
from teal.marshmallow import Phone


def test_marshmallow_color():
    class Foo(Schema):
        foo = marshmallow.Color()

    foo = Foo()
    foo.load({'foo': 'blue'})
    # todo pytest raises
    with pytest.raises(ValueError, match='\'this is not a color\' is not a recognized color.'):
        foo.load({'foo': 'this is not a color'})

    serialized = foo.dump({'foo': colour.Color('red')})
    assert serialized == {'foo': 'red'}
    serialized = foo.dump({'foo': None})
    assert serialized == {'foo': None}


def test_marshmallow_version():
    class Foo(Schema):
        foo = marshmallow.Version()

    foo = Foo()
    foo.load({'foo': '1.0.1'})
    with pytest.raises(ValueError, match='invalid version number \'this is not a version\''):
        foo.load({'foo': 'this is not a version'})

    serialized = foo.dump({'foo': version.StrictVersion('1.0.1')})
    assert serialized == {'foo': '1.0.1'}
    serialized = foo.dump({'foo': None})
    assert serialized == {'foo': None}


def test_marshmallow_url():
    class Foo(Schema):
        foo = marshmallow.URL()

    foo = Foo()
    foo.load({'foo': 'https://foo.com/bar'})
    foo.load({'foo': 'https://foo.com/'})
    with pytest.raises(ValueError):
        foo.load({'foo': 'this is not an url'})
    with pytest.raises(ValueError):
        foo.load({'foo': 'foo.com/'})

    serialized = foo.dump({'foo': urlutils.URL('https://foo.com/bar')})
    assert serialized == {'foo': 'https://foo.com/bar'}

    serialized = foo.dump({'foo': None})
    assert serialized == {'foo': None}

    class Bar(Schema):
        bar = marshmallow.URL(require_path=True)

    bar = Bar()
    bar.load({'bar': 'https://foo.com/bar'})
    with pytest.raises(ValueError):
        bar.load({'bar': 'https://foo.com'})
    with pytest.raises(ValueError):
        bar.load({'bar': 'https://foo.com/'})


def test_marshmallow_ip():
    class Foo(Schema):
        foo = marshmallow.IP()

    foo = Foo()
    foo.load({'foo': '192.168.1.1'})
    with pytest.raises(ValueError,
                       match='\'this is not an url\' does not appear to be an IPv4 or IPv6 address'):
        foo.load({'foo': 'this is not an url'})

    serialized = foo.dump({'foo': ipaddress.ip_address('192.168.1.1')})
    assert serialized == {'foo': '192.168.1.1'}

    serialized = foo.dump({'foo': None})
    assert serialized == {'foo': None}


def test_marshmallow_phone():
    class Foo(Schema):
        foo = Phone()

    foo = Foo()
    foo.load({'foo': '+34936666666'})
    with pytest.raises(PhoneNumberParseException):  # Phone number cannot be parsed
        foo.load({'foo': 'this is not a phone number'})

    with pytest.raises(ValueError, match='The phone number is invalid.'):
        # Phone number is invalid for the country rules
        foo.load({'foo': '+3401'})

    serialized = foo.dump({'foo': PhoneNumber('+34936666666')})
    assert serialized == {'foo': '+34 936 66 66 66'}

    serialized = foo.dump({'foo': None})
    assert serialized == {'foo': None}



from flask import Request as _Request, current_app as app

from teal.resource import Schema


class Request(_Request):
    def get_json(self, force=False,
                 silent=False,
                 cache=True,
                 validate=True,
                 schema: Schema = None) -> dict:
        """
        As :meth:`flask.Request.get_json` but parsing
        the resulting json through passed-in ``schema`` (or by default
        ``g.schema``).
        """
        json = super().get_json(force, silent, cache)
        if validate:
            json = schema.load(json) if schema else app.resources[self.blueprint].schema.load(json)
        return json



import ipaddress
import re
import uuid
from distutils.version import StrictVersion
from typing import Any, Type, Union

from boltons.typeutils import classproperty
from boltons.urlutils import URL as BoltonsUrl
from flask_sqlalchemy import BaseQuery, Model as _Model, SQLAlchemy as FlaskSQLAlchemy, \
    SignallingSession
from sqlalchemy import CheckConstraint, cast, event, types
from sqlalchemy.dialects.postgresql import ARRAY, INET
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.exc import MultipleResultsFound, NoResultFound
from sqlalchemy_utils import Ltree
from werkzeug.exceptions import BadRequest, NotFound, UnprocessableEntity

from teal.utils import if_none_return_none


class ResourceNotFound(NotFound):
    # todo show id
    def __init__(self, resource: str) -> None:
        super().__init__('The {} doesn\'t exist.'.format(resource))


class MultipleResourcesFound(UnprocessableEntity):
    # todo show id
    def __init__(self, resource: str) -> None:
        super().__init__('Expected only one {} but multiple where found'.format(resource))


POLYMORPHIC_ID = 'polymorphic_identity'
POLYMORPHIC_ON = 'polymorphic_on'
INHERIT_COND = 'inherit_condition'
CASCADE = 'save-update, delete'
CASCADE_OWN = '{}, delete-orphan'.format(CASCADE)
DB_CASCADE_SET_NULL = 'SET NULL'


class Query(BaseQuery):
    def one(self):
        try:
            return super().one()
        except NoResultFound:
            raise ResourceNotFound(self._entities)
        except MultipleResultsFound:
            raise MultipleResourcesFound(self._entities)


class Model(_Model):
    # Just provide typing
    query_class = Query  # type: Type[Query]
    query = None  # type: Query

    @classproperty
    def t(cls):
        return cls.__name__


class Session(SignallingSession):
    """A SQLAlchemy session that raises better exceptions."""

    def _flush(self, objects=None):
        try:
            super()._flush(objects)
        except IntegrityError as e:
            raise DBError(e)  # This creates a suitable subclass


class SchemaSession(Session):
    """ Session that is configured to use a PostgreSQL's Schema.

    Idea from `here <https://stackoverflow.com/a/9299021>`_.
    """

    def __init__(self, db, autocommit=False, autoflush=True, **options):
        super().__init__(db, autocommit, autoflush, **options)
        SCHEMA = self.app.config['SCHEMA']
        if SCHEMA:
            self.execute('SET search_path TO {}, public'.format(SCHEMA))


class SQLAlchemy(FlaskSQLAlchemy):
    def create_session(self, options):
        """As parent's create_session but adding our SchemaSession."""
        return sessionmaker(class_=Session, db=self, **options)


class SchemaSQLAlchemy(SQLAlchemy):
    """
    Enhances :class:`flask_sqlalchemy.SQLAlchemy` by using PostgreSQL's
    schemas when creating/dropping tables.

    See :attr:`teal.config.SCHEMA` for more info.
    """

    def __init__(self, app=None, use_native_unicode=True, session_options=None, metadata=None,
                 query_class=Query, model_class=Model):
        super().__init__(app, use_native_unicode, session_options, metadata, query_class,
                         model_class)
        # The following listeners set psql's search_path to the correct
        # schema and create the schemas accordingly

        # Specifically:
        # 1. Creates the schemas and set ``search_path`` to app's config SCHEMA
        event.listen(self.metadata, 'before_create', self.create_schemas)
        # Set ``search_path`` to default (``public``)
        event.listen(self.metadata, 'after_create', self.revert_connection)
        # Set ``search_path`` to app's config SCHEMA
        event.listen(self.metadata, 'before_drop', self.set_search_path)
        # Set ``search_path`` to default (``public``)
        event.listen(self.metadata, 'after_drop', self.revert_connection)

    def _execute_for_all_tables(self, app, bind, operation, skip_tables=False):
        # todo how to pass app to our event listeners without contaminating self?
        self._app = self.get_app(app)
        super()._execute_for_all_tables(app, bind, operation, skip_tables)

    def create_schemas(self, target, connection, **kw):
        """
        Create the schemas and set the active schema.

        From `here <https://bitbucket.org/zzzeek/sqlalchemy/issues/3914/
        extend-create_all-drop_all-to-include#comment-40129850>`_.
        """
        schemas = set(table.schema for table in target.tables.values() if table.schema)
        if self._app.config['SCHEMA']:
            schemas.add(self._app.config['SCHEMA'])
        for schema in schemas:
            connection.execute('CREATE SCHEMA IF NOT EXISTS {}'.format(schema))
        self.set_search_path(target, connection)

    def set_search_path(self, _, connection, **kw):
        app = self.get_app()
        if app.config['SCHEMA']:
            connection.execute('SET search_path TO {}, public'.format(app.config['SCHEMA']))

    def revert_connection(self, _, connection, **kw):
        connection.execute('SET search_path TO public')

    def create_session(self, options):
        """As parent's create_session but adding our SchemaSession."""
        return sessionmaker(class_=SchemaSession, db=self, **options)

    def drop_schema(self, app=None, schema=None):
        """Nukes a schema and everything that depends on it."""
        app = self.get_app(app)
        schema = schema or app.config['SCHEMA']
        with self.engine.begin() as conn:
            conn.execute('DROP SCHEMA IF EXISTS {} CASCADE'.format(schema))


class StrictVersionType(types.TypeDecorator):
    """StrictVersion support for SQLAlchemy as Unicode.

    Idea `from official documentation <http://docs.sqlalchemy.org/en/
    latest/core/custom_types.html#augmenting-existing-types>`_.
    """
    impl = types.Unicode

    @if_none_return_none
    def process_bind_param(self, value, dialect):
        return str(value)

    @if_none_return_none
    def process_result_value(self, value, dialect):
        return StrictVersion(value)


class URL(types.TypeDecorator):
    """bolton's URL support for SQLAlchemy as Unicode."""
    impl = types.Unicode

    @if_none_return_none
    def process_bind_param(self, value, dialect):
        return str(value)

    @if_none_return_none
    def process_result_value(self, value, dialect):
        return BoltonsUrl(value)


class IP(types.TypeDecorator):
    """ipaddress support for SQLAlchemy as PSQL INET."""
    impl = INET

    @if_none_return_none
    def process_bind_param(self, value, dialect):
        return str(value)

    @if_none_return_none
    def process_result_value(self, value, dialect):
        return ipaddress.ip_address(value)


class UUIDLtree(Ltree):
    """This Ltree only wants UUIDs as paths elements."""

    def __init__(self, path_or_ltree: Union[Ltree, uuid.UUID]):
        """
        Creates a new Ltree. If the passed-in value is an UUID,
        it automatically generates a suitable string for Ltree.
        """
        if not isinstance(path_or_ltree, Ltree):
            if isinstance(path_or_ltree, uuid.UUID):
                path_or_ltree = self.convert(path_or_ltree)
            else:
                raise ValueError('Ltree does not accept {}'.format(path_or_ltree.__class__))
        super().__init__(path_or_ltree)

    @staticmethod
    def convert(id: uuid.UUID) -> str:
        """Transforms an uuid to a ready-to-ltree str representation."""
        return str(id).replace('-', '_')


def check_range(column: str, min=1, max=None) -> CheckConstraint:
    """Database constraint for ranged values."""
    constraint = '>= {}'.format(min) if max is None else 'BETWEEN {} AND {}'.format(min, max)
    return CheckConstraint('{} {}'.format(column, constraint))


class ArrayOfEnum(ARRAY):
    """
    Allows to use Arrays of Enums for psql.

    From `the docs <http://docs.sqlalchemy.org/en/latest/dialects/
    postgresql.html?highlight=array#postgresql-array-of-enum>`_
    and `this issue <https://bitbucket.org/zzzeek/sqlalchemy/issues/
    3467/array-of-enums-does-not-allow-assigning>`_.
    """

    def bind_expression(self, bindvalue):
        return cast(bindvalue, self)

    def result_processor(self, dialect, coltype):
        super_rp = super(ArrayOfEnum, self).result_processor(
            dialect, coltype)

        def handle_raw_string(value):
            inner = re.match(r'^{(.*)}$', value).group(1)
            return inner.split(',') if inner else []

        def process(value):
            if value is None:
                return None
            return super_rp(handle_raw_string(value))

        return process


class DBError(BadRequest):
    """An Error from the database.

    This helper error is used to map SQLAlchemy's IntegrityError
    to more precise errors (like UniqueViolation) that are understood
    as a client-ready HTTP Error.

    When instantiating the class it auto-selects the best error.
    """

    def __init__(self, origin: IntegrityError):
        super().__init__(str(origin))
        self.origin = origin

    def __new__(cls, origin: IntegrityError) -> Any:
        msg = str(origin)
        if 'unique constraint' in msg.lower():
            return super().__new__(UniqueViolation)
        return super().__new__(cls)


class UniqueViolation(DBError):
    pass



from typing import Dict, Set, Type

from boltons.typeutils import issubclass
from boltons.urlutils import URL

from teal.resource import Resource


class Config:
    """
    The configuration class.

    Subclass and set here your config values.
    """
    RESOURCE_DEFINITIONS = set()  # type: Set[Type[Resource]]
    """
    A list of resource definitions to load.
    """

    SQLALCHEMY_DATABASE_URI = None  # type: str
    """
    The access to the main Database.
    """
    SQLALCHEMY_BINDS = {}  # type: Dict[str, str]
    """
    Optional extra databases. See `here <http://flask-sqlalchemy.pocoo.org
    /2.3/binds/#referring-to-binds>`_ how bind your models to different
    databases.
    """
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    """
    Disables flask-sqlalchemy notification system. 
    Save resources and hides a warning by flask-sqlalchemy itself.
    
    See `this answer in Stackoverflow for more info
    <https://stackoverflow.com/a/33790196>`_. 
    """

    SCHEMA = None  # type: str
    """
    A string describing the main PostgreSQL's schema. ``None`` disables
    this functionality.
    
    If you use a factory of apps (for example by using
    :func:`teal.teal.prefixed_database_factory`) and then set this
    value differently per each app (as each app has a separate config)
    you effectively create a `multi-tenant app <https://
    news.ycombinator.com/item?id=4268792>`_.
    
    Your models by default will be created in this ``SCHEMA``,
    unless you set something like::
    
        class User(db.Model):
            __table_args__ = {'schema': 'users'}
            
    In which case this will be created in the ``users`` schema.
    
    Schemas are interesting over having multiple databases (i.e. using
    flask-sqlalchemy's data binding) because you can have relationships
    between them.
    
    Note that this only works with PostgreSQL.
    """

    API_DOC_CONFIG_TITLE = 'Teal'
    API_DOC_CONFIG_VERSION = '0.1'
    """
    Configuration options for the api docs. They are the parameters
    passed to `apispec <http://apispec.readthedocs.io/en/
    latest/api_core.html#apispec.APISpec>`_. Prefix the configuration
    names with ``API_DOC_CONFIG_``.
    """
    API_DOC_CLASS_DISCRIMINATOR = None
    """
    Configuration options for the api docs class definitions.
    
    You can pass any `schema definition <https://github.com/OAI/
    OpenAPI-Specification/blob/master/versions/2.0.md#schemaObject>`_
    prefiex by ``API_DOC_CLASS_`` like in the example above.
    """

    CORS_ORIGINS = '*'
    CORS_EXPOSE_HEADERS = 'Authorization'
    CORS_ALLOW_HEADERS = 'Content-Type', 'Authorization'
    """
    Configuration for CORS. See the options you can pass by in `Flask-Cors 
    <https://flask-cors.corydolphin.com/en/latest/api.html#extension>`_,
    exactly in **Parameters**, like the ones above.
    """

    def __init__(self, db: str = None) -> None:
        """
        :param db: Optional. Set the ``SQLALCHEMY_DATABASE_URI`` param.
        """
        for r in self.RESOURCE_DEFINITIONS:
            assert issubclass(r, Resource), '{0!r} is not a subclass of Resource'.format(r)
        if db:
            assert URL(db), 'Set a valid URI'
            self.SQLALCHEMY_DATABASE_URI = db



import inspect
from functools import wraps
from typing import Dict, Iterator, Tuple

from sqlalchemy.dialects import postgresql
from teal import resource


def compiled(Model, query) -> Tuple[str, Dict[str, str]]:
    """
    Generates a SQL statement.

    :return A tuple with 1. the SQL statement and 2. the params for it.
    """
    c = Model.query.filter(*query).statement.compile(dialect=postgresql.dialect())
    return str(c), c.params


def import_resource(module) -> Iterator['resource.Resource']:
    """
    Gets the resource classes from the passed-in module.

    This method yields subclasses of :class:`teal.resource.Resource`
    found in the given module.
    """

    for obj in vars(module).values():
        if inspect.isclass(obj) \
                and issubclass(obj, resource.Resource) \
                and obj != resource.Resource:
            yield obj


def if_none_return_none(f):
    """If the first value is None return None, otherwise execute f."""

    @wraps(f)
    def wrapper(self, value, *args, **kwargs):
        if value is None:
            return None
        return f(self, value, *args, **kwargs)

    return wrapper



from unittest.mock import MagicMock, call

from teal import utils
from teal.resource import Resource


def test_import_resource():
    """Tests that only subclasses of resource are imported."""

    class module:
        class Foo(Resource):
            pass

        class Bar(Resource):
            pass

        class ThisIsNotResource:
            pass

        RandomVar = 3
        Res = Resource

    x = set(utils.import_resource(module))
    assert x == {module.Foo, module.Bar}


def test_if_none_return_none():
    mocked = MagicMock()
    wrapped = utils.if_none_return_none(mocked)
    assert mocked.call_count == 0
    wrapped(None, 3)
    assert mocked.call_count == 1
    assert mocked.call_args == call(None, 3)
    wrapped(None, None)
    assert mocked.call_count == 1



from teal import enums


def test_enum_currency():
    assert enums.Currency.EGP.value == 31


def test_enum_continent():
    assert enums.Continent.AF.value == 'Africa'


def test_enum_country():
    assert enums.Country.PH.value == 'Philippines'


def test_enum_subdivision():
    assert enums.Subdivision['ES-CA'].value == 963


def test_enum_subdivision_in_country():
    assert enums.Subdivision['ES-CA'] in enums.Country.ES
    assert enums.Subdivision['ES-CA'] not in enums.Country.PH



from marshmallow.fields import Integer, Nested, Str

from teal.query import Between, Equal, ILike, Or, Query, Sort, SortField, Join
from teal.teal import Teal
from teal.utils import compiled


def test_query(app: Teal):
    with app.app_context():
        Device = app.resources['Device'].MODEL

        class Q(Query):
            idq = Or(Equal(Device.id, Str()))
            modelq = ILike(Device.model)

        schema = Q()
        query = schema.load({'idq': ['a', 'b', 'c'], 'modelq': 'foobar'})
        s, params = compiled(Device, query)
        # Order between query clauses can change
        assert 'device.model ILIKE %(model_1)s' in s
        assert '(device.id = %(id_1)s OR device.id = %(id_2)s OR device.id = %(id_3)s' in s
        assert params == {'id_2': 'b', 'id_3': 'c', 'id_1': 'a', 'model_1': 'foobar%'}


def test_query_join(app: Teal):
    """Checks that nested queries work."""
    with app.app_context():
        Device = app.resources['Device'].MODEL
        Computer = app.resources['Computer'].MODEL

        class Inner(Query):
            foo = ILike(Device.model)

        class Q(Query):
            idq = Between(Device.id, Integer())
            componentq = Join(Device.id == Computer.id, Inner)

        schema = Q()
        query = schema.load({'idq': [1, 4], 'componentq': {'foo': 'bar'}})
        s, params = compiled(Device, query)
        assert 'device.id BETWEEN %(id_1)s AND %(id_2)s' in s
        assert 'device.model ILIKE %(model_1)s' in s
        assert params == {'id_1': 1, 'model_1': 'bar%', 'id_2': 4}


def test_query_sort(app: Teal):
    """Tests sorting params."""
    with app.app_context():
        Device = app.resources['Device'].MODEL

        class Sorting(Sort):
            models = SortField(Device.model)
            ids = SortField(Device.id)

        schema = Sorting()
        sort = tuple(str(s) for s in schema.load({'models': True, 'ids': False}))
        assert len(sort) == 2
        assert 'device.model ASC' in sort
        assert 'device.id DESC' in sort



from base64 import b64encode
from unittest.mock import MagicMock

from flask.json import jsonify
from flask_sqlalchemy import SQLAlchemy
from werkzeug.exceptions import Unauthorized

from teal.auth import TokenAuth
from teal.config import Config
from teal.resource import Resource, Schema, View
from teal.teal import Teal


def test_token_auth_view(db: SQLAlchemy):
    """
    Ensures that an authorization endpoint correctly protects against
    wrong credentials (this case tokens), allowing the endpoint
    to only specific cases.
    """

    class TestTokenAuth(TokenAuth):
        authenticate = MagicMock(side_effect=Unauthorized)

    class FooSchema(Schema):
        pass

    class FooView(View):
        get = MagicMock(side_effect=lambda id: jsonify({'did': 'it!'}))

    class Foo(db.Model):
        id = db.Column(db.Integer, primary_key=True)

    class FooDef(Resource):
        SCHEMA = FooSchema
        VIEW = FooView
        MODEL = Foo
        AUTH = True

    class TestConfig(Config):
        RESOURCE_DEFINITIONS = [FooDef]

    app = Teal(config=TestConfig(), Auth=TestTokenAuth, db=db)
    client = app.test_client()

    # No token
    # No auth header sent
    client.get(res=FooSchema.t, status=Unauthorized)
    assert TestTokenAuth.authenticate.call_count == 0

    # Wrong format
    # System couldn't parse Auth header
    client.get(res=FooSchema.t, token='this is wrong', status=Unauthorized)
    assert TestTokenAuth.authenticate.call_count == 0

    # Wrong credentials
    # System can parse credentials but they are incorrect
    client.get(res=FooSchema.t, token=b64encode(b'nok:').decode(), status=Unauthorized)
    # Authenticate method was hit
    assert TestTokenAuth.authenticate.call_count == 1

    # OK
    # Our authenticate method now returns some dummy user instead of
    # raising Unauthorized
    TestTokenAuth.authenticate = MagicMock(return_value={'id': '1'})
    data, _ = client.get(res=FooSchema.t, token=b64encode(b'ok:').decode())
    TestTokenAuth.authenticate.assert_called_once_with('ok', '')
    # The endpoint was hit
    assert data == {'did': 'it!'}
    FooView.get.assert_called_once_with(id=None)



from enum import Enum
from typing import Callable, Iterable, Iterator, Tuple, Type, Union

from anytree import PreOrderIter
from boltons.typeutils import classproperty, issubclass
from ereuse_utils.naming import Naming
from flask import Blueprint, current_app, g, request, url_for
from flask.json import jsonify
from flask.views import MethodView
from marshmallow import Schema as MarshmallowSchema, SchemaOpts as MarshmallowSchemaOpts, \
    ValidationError, post_dump, pre_load, validates_schema
from werkzeug.exceptions import MethodNotAllowed

from teal import db, query


class SchemaOpts(MarshmallowSchemaOpts):
    """
    Subclass of Marshmallow's SchemaOpts that provides
    options for Teal's schemas.
    """

    def __init__(self, meta, ordered=False):
        super().__init__(meta, ordered)
        self.PREFIX = meta.PREFIX


class Schema(MarshmallowSchema):
    """
    The definition of the fields of a resource.
    """
    OPTIONS_CLASS = SchemaOpts

    class Meta:
        PREFIX = None
        """Optional. A prefix for the type; ex. devices:Computer."""

    # noinspection PyMethodParameters
    @classproperty
    def t(cls: Type['Schema']) -> str:
        """The type for this schema, auto-computed from its name."""
        name, *_ = cls.__name__.split('Schema')
        return Naming.new_type(name, cls.Meta.PREFIX)

    # noinspection PyMethodParameters
    @classproperty
    def resource(cls: Type['Schema']) -> str:
        """The resource name of this schema."""
        return Naming.resource(cls.t)

    @validates_schema(pass_original=True)
    def check_unknown_fields(self, _, original_data: dict):
        """
        Raises a validationError when user sends extra fields.

        From `Marshmallow docs<http://marshmallow.readthedocs.io/en/
        latest/extending.html#validating-original-input-data>`_.
        """
        unknown_fields = set(original_data) - set(f.data_key or n for n, f in self.fields.items())
        if unknown_fields:
            raise ValidationError('Unknown field', unknown_fields)

    @validates_schema(pass_original=True)
    def check_dump_only(self, _, orig_data: dict):
        """
        Raises a ValidationError if the user is submitting
        'read-only' fields.
        """
        # Note that validates_schema does not execute when dumping
        dump_only_fields = (name for name, field in self.fields.items() if field.dump_only)
        non_writable = set(orig_data).intersection(dump_only_fields)
        if non_writable:
            raise ValidationError('Non-writable field', non_writable)

    @pre_load
    @post_dump
    def remove_none_values(self, data: dict) -> dict:
        """
        Skip from dumping and loading values that are None.

        A value that is None will be the same as a value that has not
        been set.

        `From here <https://github.com/marshmallow-code/marshmallow/
        issues/229#issuecomment-134387999>`_.
        """
        # Will I always want this?
        # maybe this could be a setting in the future?
        return {key: value for key, value in data.items() if value is not None}

    def dump(self,
             model: Union['db.Model', Iterable['db.Model']],
             many=None,
             update_fields=True,
             nested=None,
             polymorphic_on='t'):
        """
        Like marshmallow's dump but with nested resource support and
        it only works for Models.

        This can load model relationships up to ``nested`` level. For
        example, if ``nested`` is ``1`` and we pass in a model of
        ``User`` that has a relationship with a table of ``Post``, it
        will load ``User`` and ``User.posts`` with all posts objects
        populated, but it won't load relationships inside the
        ``Post`` object. If, at the same time the ``Post`` has
        an ``author`` relationship with ``author_id`` being the FK,
        ``user.posts[n].author`` will be the value of ``author_id``.

        Define nested fields with the :class:`teal.marshmallow.NestedOn`

        This method requires an active application context as it needs
        to store some stuff in ``g``.

        :param nested: How many layers of nested relationships to load?
                       By default only loads 1 nested relationship.
        """
        from teal.marshmallow import NestedOn
        if nested is not None:
            setattr(g, NestedOn.NESTED_LEVEL, 0)
            setattr(g, NestedOn.NESTED_LEVEL_MAX, nested)
        if many:
            # todo this breaks with normal dicts. Maybe this should go
            # in NestedOn in the same way it happens when loading
            if isinstance(model, dict):
                return super().dump(model, update_fields=update_fields)
            else:
                return [self._polymorphic_dump(o, update_fields, polymorphic_on) for o in model]

        else:
            if isinstance(model, dict):
                return super().dump(model, update_fields=update_fields)
            else:
                return self._polymorphic_dump(model, update_fields, polymorphic_on)

    def _polymorphic_dump(self, obj: 'db.Model', update_fields, polymorphic_on='t'):
        schema = current_app.resources[getattr(obj, polymorphic_on)].schema
        if schema.t != self.t:
            return super(schema.__class__, schema).dump(obj, False, update_fields)
        else:
            return super().dump(obj, False, update_fields)

    def jsonify(self,
                model: Union['db.Model', Iterable['db.Model']],
                nested=1,
                many=False,
                update_fields: bool = True,
                polymorphic_on='t',
                **kw) -> str:
        """
        Like flask's jsonify but with model / marshmallow schema
        support.

        :param nested: How many layers of nested relationships to load?
                       By default only loads 1 nested relationship.
        """
        return jsonify(self.dump(model, many, update_fields, nested, polymorphic_on))


class View(MethodView):
    """
    A REST interface for resources.
    """
    QUERY_PARSER = query.NestedQueryFlaskParser()

    class FindArgs(MarshmallowSchema):
        """
        Allowed arguments for the ``find``
        method (GET collection) endpoint
        """

    def __init__(self, definition: 'Resource', **kw) -> None:
        self.resource_def = definition
        """The ResourceDefinition tied to this view."""
        self.schema = None  # type: Schema
        """The schema tied to this view."""
        self.find_args = self.FindArgs()
        super().__init__()

    def dispatch_request(self, *args, **kwargs):
        # This is unique for each view call
        self.schema = g.schema
        """
        The default schema in this resource. 
        Added as an attr for commodity; you can always use g.schema.
        """
        return super().dispatch_request(*args, **kwargs)

    def get(self, id):
        """Get a collection of resources or a specific one.
        ---
        parameters:
        - name: id
          in: path
          description: The identifier of the resource.
          type: string
          required: false
        responses:
          200:
            description: Return the collection or the specific one.
        """
        if id:
            response = self.one(id)
        else:
            args = self.QUERY_PARSER.parse(self.find_args,
                                           request,
                                           locations=('querystring',))
            response = self.find(args)
        return response

    def one(self, id):
        """GET one specific resource (ex. /cars/1)."""
        raise MethodNotAllowed()

    def find(self, args: dict):
        """GET a list of resources (ex. /cars)."""
        raise MethodNotAllowed()

    def post(self):
        raise MethodNotAllowed()

    def delete(self, id):
        raise MethodNotAllowed()

    def put(self, id):
        raise MethodNotAllowed()

    def patch(self, id):
        raise MethodNotAllowed()


class Converters(Enum):
    """An enumeration of available URL converters."""
    string = 'string'
    int = 'int'
    float = 'float'
    path = 'path'
    any = 'any'
    uuid = 'uuid'


class Resource(Blueprint):
    """
    Main resource class. Defines the schema, views,
    authentication, database and collection of a resource.

    A ``ResourceDefinition`` is a Flask
    :class:`flask.blueprints.Blueprint` that provides everything
    needed to set a REST endpoint.
    """
    VIEW = None  # type: Type[View]
    """
    Resource view linked to this definition or None.
    If none, this resource does not generate any view.
    """
    SCHEMA = Schema  # type: Type[Schema]
    """The Schema that validates a submitting resource at the entry point."""
    AUTH = False
    """
    If true, authentication is required for all the endpoints of this 
    resource defined in ``VIEW``.
    """
    ID_NAME = 'id'
    """
    The variable name for GET *one* operations that is used as an id.
    """
    ID_CONVERTER = Converters.string
    """
    The converter for the id.

    Note that converters do **cast** the value, so the converter 
    ``uuid`` will return an ``UUID`` object.
    """
    __type__ = None  # type: str
    """
    The type of resource. 
    If none, it is used the type of the Schema (``Schema.type``)
    """

    def __init__(self, app,
                 import_name=__name__.split('.')[0],
                 static_folder=None,
                 static_url_path=None,
                 template_folder=None,
                 url_prefix=None,
                 subdomain=None,
                 url_defaults=None,
                 root_path=None,
                 cli_commands: Iterable[Tuple[Callable, str or None]] = tuple()):
        assert not self.VIEW or issubclass(self.VIEW, View), 'VIEW should be an subclass of View'
        assert issubclass(self.SCHEMA, Schema), 'SCHEMA should be a subclass of Schema'
        url_prefix = url_prefix if url_prefix is not None else '/{}'.format(self.resource)
        super().__init__(self.type, import_name, static_folder, static_url_path, template_folder,
                         url_prefix, subdomain, url_defaults, root_path)
        self.app = app
        self.schema = self.SCHEMA()
        # Views
        if self.VIEW:
            view = self.VIEW.as_view('main', definition=self, auth=app.auth)
            if self.AUTH:
                view = app.auth.requires_auth(view)
            self.add_url_rule('/', defaults={'id': None}, view_func=view, methods={'GET'})
            self.add_url_rule('/', view_func=view, methods={'POST'})
            self.add_url_rule('/<{}:{}>'.format(self.ID_CONVERTER.value, self.ID_NAME),
                              view_func=view, methods={'GET', 'PUT', 'DELETE'})
        self.cli_commands = cli_commands
        self.before_request(self.load_resource)

    @classproperty
    def type(cls):
        return cls.__type__ or cls.SCHEMA.t

    @classproperty
    def t(cls):
        return cls.type

    @classproperty
    def resource(cls):
        return Naming.resource(cls.type)

    def load_resource(self):
        """
        Loads a schema and resource_def into the current request so it
        can be used easily by functions outside view.
        """
        g.schema = self.schema
        g.resource_def = self

    def init_db(self, db: 'db.SQLAlchemy'):
        """
        Put here code to execute when initializing the database for this
        resource.

        We guarantee this to be executed in an app_context.

        No need to commit.
        """
        pass

    @property
    def subresources_types(self) -> Iterator[str]:
        """Gets the types of the subresources."""
        return (node.name for node in PreOrderIter(self.app.tree[self.t]))


def url_for_resource(
        resource: Union[Resource, Schema, 'db.Model', str, Type[Resource],
                        Type[Schema], Type['db.Model']],
        item_id=None) -> str:
    """
    As Flask's ``url_for``, this generates an URL but specifically for
    a View endpoint of the given resource.
    :param resource:
    :param item_id: If given, append the ID of the resource in the URL,
                    ex. GET /devices/1
    :return: An URL.
    """
    type = getattr(resource, 't', resource)
    values = {}
    if item_id:
        values[current_app.resources[type].ID_NAME] = item_id
    return url_for('{}.main'.format(type), **values)



import ipaddress
from distutils.version import StrictVersion
from typing import Type, Union

import colour
from boltons import urlutils
from flask import current_app as app, g
from marshmallow import utils
from marshmallow.fields import Field, Nested as MarshmallowNested, \
    ValidationError as _ValidationError, missing_
from marshmallow.validate import Validator
from marshmallow_enum import EnumField as _EnumField
from sqlalchemy_utils import PhoneNumber

from teal import db as tealdb
from teal.resource import Schema
from teal.utils import if_none_return_none


class Version(Field):
    """A python StrictVersion field, like '1.0.1'."""

    @if_none_return_none
    def _serialize(self, value, attr, obj):
        return str(value)

    @if_none_return_none
    def _deserialize(self, value, attr, data):
        return StrictVersion(value)


class Color(Field):
    """Any color field that can be accepted by the colour package."""

    @if_none_return_none
    def _serialize(self, value, attr, obj):
        return str(value)

    @if_none_return_none
    def _deserialize(self, value, attr, data):
        return colour.Color(value)


class URL(Field):
    def __init__(self, require_path=False,
                 default=missing_, attribute=None, data_key=None, error=None, validate=None,
                 required=False, allow_none=None, load_only=False, dump_only=False,
                 missing=missing_, error_messages=None, **metadata):
        super().__init__(default, attribute, data_key, error, validate, required, allow_none,
                         load_only, dump_only, missing, error_messages, **metadata)
        self.require_path = require_path

    @if_none_return_none
    def _serialize(self, value, attr, obj):
        return value.to_text()

    @if_none_return_none
    def _deserialize(self, value, attr, data):
        url = urlutils.URL(value)
        if url.scheme or url.host:
            if self.require_path:
                if url.path and url.path != '/':
                    return url
            else:
                return url
        raise ValueError('Not a valid URL.')


class IP(Field):
    @if_none_return_none
    def _serialize(self, value: Union[ipaddress.IPv4Address, ipaddress.IPv6Address], attr, obj):
        return str(value)

    @if_none_return_none
    def _deserialize(self, value: str, attr, data):
        return ipaddress.ip_address(value)


class Phone(Field):
    @if_none_return_none
    def _serialize(self, value: PhoneNumber, attr, obj):
        return value.international

    @if_none_return_none
    def _deserialize(self, value: str, attr, data):
        phone = PhoneNumber(value)
        if not phone.is_valid_number():
            raise ValueError('The phone number is invalid.')
        return phone


class NestedOn(MarshmallowNested):
    """A relationship with a resource schema that emulates the
    relationships in SQLAlchemy.

    It allows instantiating SQLA models when deserializing NestedOn
    values in two fashions:

    - If the :attr:`.only_query` is set, NestedOn expects a scalar
      (str, int...) value when deserializing, and tries to get
      an existing model that has such value. Typical case is setting
      :attr:`.only_query` to ``id``, and then pass-in the id
      of a nested model. In such case NestedOn will change the id
      for the model representing the ID.
    - If :attr:`.only_query` is not set, NestedOn expects the
      value to deserialize to be a dictionary, and instantiates
      the model with the values of the dictionary. In this case
      NestedOn requires :attr:`.polymorphic_on` to be set as a field,
      usually called ``type``, that references a subclass of Model;
      ex. {'type': 'SpecificDevice', ...}.

    When serializing from :meth:`teal.resource.Schema.jsonify` it
    serializes nested relationships up to a defined limit.

    :param polymorphic_on: The field name that discriminates
                               the type of object. For example ``type``.
                               Then ``type`` contains the class name
                               of a subschema of ``nested``.
    """
    NESTED_LEVEL = '_level'
    NESTED_LEVEL_MAX = '_level_max'

    def __init__(self,
                 nested,
                 polymorphic_on: str,
                 db: tealdb.SQLAlchemy,
                 collection_class=list,
                 default=missing_,
                 exclude=tuple(),
                 only_query: str = None,
                 only=None,
                 **kwargs):
        self.polymorphic_on = polymorphic_on
        self.collection_class = collection_class
        self.only_query = only_query
        assert isinstance(polymorphic_on, str)
        assert isinstance(only, str) or only is None
        super().__init__(nested, default, exclude, only, **kwargs)
        self.db = db

    def _deserialize(self, value, attr, data):
        if self.many and not utils.is_collection(value):
            self.fail('type', input=value, type=value.__class__.__name__)

        if isinstance(self.only, str):  # self.only is a field name
            if self.many:
                value = self.collection_class({self.only: v} for v in value)
            else:
                value = {self.only: value}
        # New code:
        parent_schema = app.resources[super().schema.t].SCHEMA
        if self.many:
            return self.collection_class(self._deserialize_one(single, parent_schema, attr)
                                         for single in value)
        else:
            return self._deserialize_one(value, parent_schema, attr)

    def _deserialize_one(self, value, parent_schema: Type[Schema], attr):
        if isinstance(value, dict) and self.polymorphic_on in value:
            type = value[self.polymorphic_on]
            resource = app.resources[type]
            if not issubclass(resource.SCHEMA, parent_schema):
                raise ValidationError('{} is not a sub-type of {}'.format(type, parent_schema.t),
                                      field_names=[attr])
            schema = resource.SCHEMA(only=self.only,
                                     exclude=self.exclude,
                                     context=getattr(self.parent, 'context', {}),
                                     load_only=self._nested_normalized_option('load_only'),
                                     dump_only=self._nested_normalized_option('dump_only'))
            schema.ordered = getattr(self.parent, 'ordered', False)
            value = schema.load(value)
            model = self._model(type)(**value)
        elif self.only_query:  # todo test only_query
            model = self._model(parent_schema.t).query.filter_by(**{self.only_query: value}).one()
        else:
            raise ValidationError('\'Type\' field required to disambiguate resources.',
                                  field_names=[attr])
        assert isinstance(model, tealdb.Model)
        return model

    def _model(self, type: str) -> Type[tealdb.Model]:
        """Given the type of a model it returns the model class."""
        return self.db.Model._decl_class_registry.data[type]()

    def serialize(self, attr: str, obj, accessor=None) -> dict:
        """See class docs."""
        if g.get(NestedOn.NESTED_LEVEL) == g.get(NestedOn.NESTED_LEVEL_MAX):
            # Idea from https://marshmallow-sqlalchemy.readthedocs.io
            # /en/latest/recipes.html#smart-nested-field
            # Gets the FK of the relationship instead of the full object
            # This won't work for many-many relationships (as they are lists)
            # In such case return None
            # todo is this the behaviour we want?
            return getattr(obj, attr + '_id', None)
        setattr(g, NestedOn.NESTED_LEVEL, g.get(NestedOn.NESTED_LEVEL) + 1)
        ret = super().serialize(attr, obj, accessor)
        setattr(g, NestedOn.NESTED_LEVEL, g.get(NestedOn.NESTED_LEVEL) - 1)
        return ret


class IsType(Validator):
    """
    Validator which succeeds if the value it is passed is a registered
    resource type.

    :param parent: If set, type must be a subtype of such resource.
                   By default accept any resource.
    """
    # todo remove if not needed
    no_type = 'Type does not exist.'
    no_subtype = 'Type is not a descendant type of {parent}'

    def __init__(self, parent: str = None) -> None:
        self.parent = parent  # type: str

    def _repr_args(self):
        return 'parent={0!r}'.format(self.parent)

    def __call__(self, type: str):
        assert not self.parent or self.parent in app.resources
        try:
            r = app.resources[type]
            if self.parent:
                if not issubclass(r.__class__, app.resources[self.parent].__class__):
                    raise ValidationError(self.no_subtype.format(self.parent))
        except KeyError:
            raise ValidationError(self.no_type)


class ValidationError(_ValidationError):
    code = 422


class EnumField(_EnumField):
    """
    An EnumField that allows
    generating OpenApi enums through Apispec.
    """

    def __init__(self, enum, by_value=False, load_by=None, dump_by=None, error='', *args,
                 **kwargs):
        super().__init__(enum, by_value, load_by, dump_by, error, *args, **kwargs)
        self.metadata['enum'] = [e.name for e in enum]



"""
Tests resources on the app level
"""
from typing import Tuple
from unittest.mock import MagicMock

import pytest
from flask import Response, request
from flask.json import jsonify
from flask_sqlalchemy import SQLAlchemy
from marshmallow.fields import Integer
from werkzeug.exceptions import MethodNotAllowed, NotFound, UnprocessableEntity

from teal.client import Client
from teal.config import Config
from teal.marshmallow import IsType, ValidationError
from teal.resource import Resource as ResourceDef
from teal.teal import Teal
from tests.conftest import populated_db


def test_init(app: Teal):
    """
    Tests the initialization of resources with their relationships
    and inheritance.
    """
    DeviceDef, ComponentDef, ComputerDef = \
        app.config['RESOURCE_DEFINITIONS']  # type: Tuple[ResourceDef]
    assert isinstance(app.resources['Device'], DeviceDef)
    assert isinstance(app.resources['Computer'], ComputerDef)
    assert isinstance(app.resources['Component'], ComponentDef)
    assert app.tree['Device'].parent is None
    assert app.tree['Computer'] in app.tree['Device'].descendants
    assert app.tree['Component'] in app.tree['Device'].descendants
    assert len(app.tree['Device'].descendants) == 2
    assert app.tree['Computer'].parent == app.tree['Component'].parent == app.tree['Device']

    views = {
        'Component.main',
        'Computer.main',
        'Device.main',
        'apidocs_endpoint',
        'static'
    }
    assert views == set(app.view_functions.keys())


def test_models(app: Teal, db: SQLAlchemy):
    """Checks that the models used in the fixture work."""
    DeviceDef, ComponentDef, ComputerDef = \
        app.config['RESOURCE_DEFINITIONS']  # type: Tuple[ResourceDef]
    Component = ComponentDef.MODEL
    Computer = ComputerDef.MODEL

    component = Component(id=1, model='foo')
    pc = Computer(id=2, model='bar', components=[component])
    with app.app_context():
        db.session.add(pc)
        queried_pc = Computer.query.first()
        assert pc == queried_pc


def test_validator_is_type(app: Teal):
    """Checks the validator IsType"""
    with app.app_context():
        is_type = IsType()
        is_type('Device')
        is_type('Component')
        with pytest.raises(ValidationError):
            is_type('Foo')

        is_subtype = IsType('Component')
        is_subtype('Component')
        with pytest.raises(ValidationError):
            is_subtype('Computer')


def test_inheritance_access(fconfig: Config, db: SQLAlchemy):
    """
    Tests that the right endpoint is called when accessing sub-resources.
    """
    DeviceDef, ComponentDef, ComputerDef = fconfig.RESOURCE_DEFINITIONS  # type: Tuple[ResourceDef]

    DUMMY_DICT = {'ok': 'yes'}

    def foo(*args, **kw):
        return jsonify(DUMMY_DICT)

    DeviceDef.VIEW.get = MagicMock(side_effect=foo)
    ComponentDef.VIEW.get = MagicMock(side_effect=foo)
    ComputerDef.VIEW.get = MagicMock(side_effect=foo)

    client = Teal(config=fconfig, db=db).test_client()  # type: Client

    # Access any non-defined URI
    client.get(uri='/this-does-not-exist', status=NotFound)
    assert DeviceDef.VIEW.get.call_count == \
           ComponentDef.VIEW.get.call_count == \
           ComputerDef.VIEW.get.call_count == 0

    # Access to a non-defined method for a resource
    client.post(res=DeviceDef.type, status=MethodNotAllowed, data=dict())
    assert DeviceDef.VIEW.get.call_count == \
           ComponentDef.VIEW.get.call_count == \
           ComputerDef.VIEW.get.call_count == 0

    # Get top resource Device
    # Only device endpoint is called
    d, _ = client.get(res=DeviceDef.type)
    assert d == DUMMY_DICT
    DeviceDef.VIEW.get.assert_called_once_with(id=None)
    assert ComponentDef.VIEW.get.call_count == 0
    assert ComputerDef.VIEW.get.call_count == 0

    # Get computer
    # Only component endpoint is called
    d, _ = client.get(res=ComputerDef.type)
    assert d == DUMMY_DICT
    assert DeviceDef.VIEW.get.call_count == 1  # from before
    assert ComponentDef.VIEW.get.call_count == 0
    ComputerDef.VIEW.get.assert_called_once_with(id=None)


def test_post(fconfig: Config, db: SQLAlchemy):
    """
    Tests posting resources, going through API (Marshmallow) and DB
    (SQLAlchemy) validation, and retrieving and returning a result.
    """
    DeviceDef, ComponentDef, ComputerDef = fconfig.RESOURCE_DEFINITIONS  # type: Tuple[ResourceDef]
    Computer = ComputerDef.MODEL
    Component = ComponentDef.MODEL
    PC = {
        'id': 1,
        'model': 'foo',
        'components': [{'id': 2, 'type': 'Component'}, {'id': 3, 'type': 'Component'}]
    }

    def post():
        pc = request.get_json()
        pc = Computer(**pc)
        db.session.add(pc)
        db.session.commit()
        return Response(status=201)

    def _one(id):
        pc = Computer.query.filter_by(id=id).first()
        return_pc = {
            'id': pc.id,
            'model': pc.model,
            'type': pc.type,
        }
        # todo convert components to JSON
        return return_pc

    def one(id: int):
        return jsonify(_one(id))

    def find(_):
        return jsonify([_one(1)])

    ComputerDef.VIEW.post = MagicMock(side_effect=post)
    ComputerDef.VIEW.one = MagicMock(side_effect=one)
    ComputerDef.VIEW.find = MagicMock(side_effect=find)

    app = Teal(config=fconfig, db=db)

    client = app.test_client()  # type: Client
    with populated_db(db, app):
        client.post(res=ComputerDef.type, data=PC)
        # Wrong data
        data, _ = client.post(res=ComputerDef.type, data={'id': 'foo'}, status=ValidationError)
        assert data == {
            'code': 422,
            'type': 'ValidationError',
            'message': {'id': ['Not a valid integer.']}
        }
        # Get the first data
        data, _ = client.get(res=ComputerDef.type, item=1)
        assert data == {'id': 1, 'model': 'foo', 'type': 'Computer'}
        # Get all data
        data, _ = client.get(res=ComputerDef.type)
        assert data == [{'id': 1, 'model': 'foo', 'type': 'Computer'}]


def test_item_path(fconfig: Config, db: SQLAlchemy):
    """
    Tests that the URL converter works for the item endpoints of
    the resources.

    The following test has set an URL converter of type int, and will
    allow only integers using the flask rules.
    """
    DeviceDef, *_ = fconfig.RESOURCE_DEFINITIONS  # type: Tuple[ResourceDef, ...]

    def cannot_find(id):
        assert id == 1
        return Response(status=200)

    DeviceDef.VIEW.one = MagicMock(side_effect=cannot_find)
    app = Teal(config=fconfig, db=db)
    client = app.test_client()  # type: Client
    with populated_db(db, app):
        # All ok, we expect an int and got an int
        client.get(res=DeviceDef.type, item=1)
        DeviceDef.VIEW.one.assert_called_once_with(1)
        # Conversion of 'int' works in here
        client.get(res=DeviceDef.type, item='1')
        assert DeviceDef.VIEW.one.call_count == 2
        # Anything else fails and our function is directly not executed
        client.get(res=DeviceDef.type, item='foo', status=NotFound)
        assert DeviceDef.VIEW.one.call_count == 2


def test_args(fconfig: Config, db: SQLAlchemy):
    """Tests the handling of query arguments in the URL."""
    DeviceDef, *_ = fconfig.RESOURCE_DEFINITIONS  # type: Tuple[ResourceDef]

    class FindArgsFoo(DeviceDef.VIEW.FindArgs):
        foo = Integer()

    DeviceDef.VIEW.FindArgs = FindArgsFoo

    def find(args: dict):
        assert args == {'foo': 25}
        return Response(status=200)

    DeviceDef.VIEW.find = MagicMock(side_effect=find)

    client = Teal(config=fconfig, db=db).test_client()  # type: Client

    # Ok
    client.get(res=DeviceDef.type, query=[('foo', 25)])
    # Extra not needed data
    client.get(res=DeviceDef.type, query=[('foo', 25), ('bar', 'nope')])
    # Wrong data
    r, _ = client.get(res=DeviceDef.type, query=[('foo', 'nope')], status=UnprocessableEntity)
    # todo r should contain descriptive message of why it fails


def test_http_exception(fconfig: Config, db: SQLAlchemy):
    """Tests correct handling of HTTP exceptions."""
    DeviceDef, *_ = fconfig.RESOURCE_DEFINITIONS  # type: Tuple[ResourceDef]

    DeviceDef.VIEW.get = MagicMock(side_effect=NotFound)
    client = Teal(config=fconfig, db=db).test_client()  # type: Client
    d, _ = client.get(res=DeviceDef.type, status=NotFound)
    assert d == {
        'code': 404,
        'message': 'The requested URL was not found on the server.  '
                   'If you entered the URL manually please check your spelling and try again.',
        'type': 'NotFound'
    }



from copy import deepcopy

import pytest
from flask_sqlalchemy import SQLAlchemy
from marshmallow import ValidationError
from marshmallow.fields import Integer
from sqlalchemy import Column

from teal.auth import Auth
from teal.config import Config
from teal.db import POLYMORPHIC_ID, POLYMORPHIC_ON
from teal.resource import Resource, Schema, View, url_for_resource
from teal.teal import Teal


def test_schema_type():
    class Foo(Schema): pass

    foo = Foo()
    assert foo.t == 'Foo'

    class FooSchema(Schema): pass

    foo_schema = FooSchema()
    assert foo_schema.t == 'Foo'


def test_model_type(db: SQLAlchemy):
    TYPE = 'Foo'

    class Foo(db.Model):
        id = db.Column(db.Integer, primary_key=True)
        type = db.Column(db.String(50))

        __mapper_args__ = {
            POLYMORPHIC_ID: TYPE,
            POLYMORPHIC_ON: type
        }

    foo = Foo()
    assert foo.type == TYPE


def test_resource_def_init(db: SQLAlchemy):
    class FooSchema(Schema): pass

    class FooView(View): pass

    class Foo(db.Model):
        id = db.Column(db.Integer, primary_key=True)

    class FooDef(Resource):
        SCHEMA = FooSchema
        VIEW = FooView
        MODEL = Foo

    class MockApp:

        def __init__(self) -> None:
            self.auth = Auth()

    foos = FooDef(MockApp())
    assert foos.SCHEMA().t == foos.type == 'Foo'
    assert foos.url_prefix == '/foos'
    assert foos.name == 'Foo'


def test_schema_extra_fields():
    """Ensures that validation doesn't let extra non-defined fields."""

    class FooSchema(Schema):
        foo = Integer()

    foos = FooSchema()

    assert foos.load({'foo': 1}) == {'foo': 1}
    with pytest.raises(ValidationError):
        # var doesn't exist in the schema
        foos.load({'foo': 2, 'bar': 'no!'})


def test_url_for_resource(app: Teal):
    with app.test_request_context():
        # Note we need a test_request_context or flask won't know
        # which base_url to use.
        assert url_for_resource('Computer') == '/computers/'
        assert url_for_resource('Computer', 24) == '/computers/24'


def test_resource_without_path(config: Config, db: SQLAlchemy):
    """Test resources that don't have url_prefix."""

    class FooDef(Resource):
        VIEW = View
        __type__ = 'Foo'

        def __init__(self, app,
                     import_name=__name__.split('.')[0],
                     static_folder=None,
                     static_url_path=None,
                     template_folder=None,
                     url_prefix='',  # We set url_prefix to empty string
                     subdomain=None,
                     url_defaults=None,
                     root_path=None):
            super().__init__(app, import_name, static_folder, static_url_path, template_folder,
                             url_prefix, subdomain, url_defaults, root_path)

    config.RESOURCE_DEFINITIONS = FooDef,

    app = Teal(config=config, db=db)
    with app.test_request_context():
        assert url_for_resource(FooDef) == '/'
        assert url_for_resource(FooDef, 1) == '/1'


def test_init_db(db: SQLAlchemy, config: Config):
    """Tests :meth:`teal.resource.Resource.init_db`."""

    class Foo(db.Model):
        id = Column(db.Integer, primary_key=True)

    class FooDef(Resource):
        def init_db(self, db: SQLAlchemy):
            db.session.add(Foo())

    config.RESOURCE_DEFINITIONS = FooDef,
    app = Teal(config=config, db=db)
    with app.app_context():
        app.init_db()
    with app.app_context():
        # If no commit happened in init_db() or anything else
        # this would not exist
        assert Foo.query.filter_by(id=1).one()

    # Test again but executing init-db through the command-line
    runner = app.test_cli_runner()
    runner.invoke(args=['init-db'], catch_exceptions=False)
    with app.app_context():
        assert Foo.query.filter_by(id=2).one()

    # Test with --erase option
    runner.invoke(args=['init-db', '--erase'], catch_exceptions=False)
    with app.app_context():
        assert Foo.query.count() == 1


def test_schema_non_writable():
    """Ensures that the user does not upload readonly fields."""

    class FooSchema(Schema):
        foo = Integer(dump_only=True)
        bar = Integer()

    foos = FooSchema()

    with pytest.raises(ValidationError, message={'id': ['Non-writable field']}):
        foos.load({'foo': 1, 'bar': 2})

    # Correctly submit without the value
    foos.load({'bar': 2})
    # Dump is not affected by this validation
    foos.dump({'foo': 1, 'bar': 2})


def test_nested_on(fconfig: Config, db: SQLAlchemy):
    """Tests the NestedOn marshmallow field."""
    DeviceDef, ComponentDef, ComputerDef = fconfig.RESOURCE_DEFINITIONS

    class GraphicCardSchema(ComponentDef.SCHEMA):
        speed = Integer()

    class GraphicCard(ComponentDef.MODEL):
        speed = db.Column(db.Integer)

    class GraphicCardDef(ComponentDef):
        SCHEMA = GraphicCardSchema
        MODEL = GraphicCard

    fconfig.RESOURCE_DEFINITIONS += (GraphicCardDef,)

    app = Teal(config=fconfig, db=db)

    pc_template = {
        'id': 1,
        'components': [
            {'id': 2, 'type': 'Component'},
            {'id': 3, 'type': 'GraphicCard', 'speed': 4}
        ]
    }
    with app.app_context():
        schema = app.resources['Computer'].SCHEMA()
        result = schema.load(pc_template)
        assert pc_template['id'] == result['id']
        assert isinstance(result['components'][0], ComponentDef.MODEL)
        assert isinstance(result['components'][1], GraphicCardDef.MODEL)
        # Let's add the graphic card's speed field to the component
        with pytest.raises(ValidationError, message={'components': {'speed': ['Unknown field']}}):
            pc = deepcopy(pc_template)
            pc['components'][0]['speed'] = 4
            schema.load(pc)
        # Let's remove the 'type'
        with pytest.raises(ValidationError,
                           message={
                               'components': ['\'Type\' field required to disambiguate resources.']
                           }):
            pc = deepcopy(pc_template)
            del pc['components'][0]['type']
            del pc['components'][1]['type']
            schema.load(pc)
        # Let's set a 'type' that is not a Component
        with pytest.raises(ValidationError,
                           message={'components': ['Computer is not a sub-type of Component']}):
            pc = deepcopy(pc_template)
            pc['components'][0]['type'] = 'Computer'
            schema.load(pc)



from teal.client import Client


def test_apispec(client: Client):
    """Tests the output of Apispec"""
    api, _ = client.get('/apidocs')
    assert 'Computer' in api['definitions']
    assert 'Component' in api['definitions']
    assert 'Device' in api['definitions']
    # Ensure resources have endpoints
    assert set(api['paths'].keys()) == {
        '/computers/',
        '/components/',
        '/devices/',
        '/apidocs'
    }



from distutils.version import StrictVersion
from typing import Tuple
from unittest.mock import MagicMock

from flask import Response, json

from teal.client import Client
from teal.config import Config
from teal.db import SchemaSQLAlchemy
from teal.teal import Teal, prefixed_database_factory


def test_prefixed_database_factory():
    """Tests using the database factory middleware."""
    dbs = ('foo', 'sqlite:////tmp/foo.db'), ('bar', 'sqlite:////tmp/bar.db')
    apps = prefixed_database_factory(Config, dbs, Teal)
    assert isinstance(apps.app, Teal)
    assert all(isinstance(app, Teal) for app in apps.mounts.values())
    # todo perform GET or something


def test_json_encoder(app: Teal):
    """
    Ensures that Teal is using the custom JSON Encoder through Flask's
    json.
    """
    with app.app_context():
        # Try to dump a type flask's json encoder cannot handle
        json.dumps({'foo': StrictVersion('1.3')})


def test_cors(fconfig: Config, db: SchemaSQLAlchemy):
    DeviceDef, *_ = fconfig.RESOURCE_DEFINITIONS  # type: Tuple[ResourceDef]

    def foo(*args, **kw):
        return Response(status=200)

    DeviceDef.VIEW.get = MagicMock(side_effect=foo)
    client = Teal(config=fconfig, db=db).test_client()  # type: Client
    _, response = client.get('/devices/')
    headers = response.headers.to_list()
    assert ('Access-Control-Expose-Headers', 'Authorization') in headers
    assert ('Access-Control-Allow-Origin', '*') in headers



from functools import wraps
from typing import Callable

from flask import current_app, g, request
from werkzeug.datastructures import Authorization
from werkzeug.exceptions import Unauthorized


class Auth:
    """
    Authentication handler for Teal.

    To authenticate the user (perform login):
    1. Set Resource.AUTH to True, or manually decorate the view with
      @auth.requires_auth
    2. Extend any subclass of this one (like TokenAuth).
    3. Implement the authenticate method with the authentication logic.
       For example, in TokenAuth here you get the user from the token.
    5. Set in your teal the Auth class you have created so
       teal can use it.
    """

    API_DOCS = {
        'type': 'http',
        'description:': 'HTTP Basic scheme',
        'name': 'Authorization',
        'in': 'header',
        'scheme': 'basic'
    }

    @classmethod
    def requires_auth(cls, f: Callable):
        """
        Decorate a view enforcing authentication (logged in user).
        """

        @wraps(f)
        def decorated(*args, **kwargs):
            auth = request.authorization
            if not auth:
                raise Unauthorized('Provide proper authorization credentials')
            current_app.auth.perform_auth(auth)
            return f(*args, **kwargs)

        return decorated

    def perform_auth(self, auth: Authorization):
        """
        Authenticate an user. This loads the user.

        An exception (expected Unauthorized) is raised if
        authentication failed.
        """
        g.user = self.authenticate(auth.username, auth.password)

    def authenticate(self, username: str, password: str) -> object:
        """
        The authentication logic. The result of this method is
        a user or a raised exception, like Werkzeug's Unauthorized,
        if authentication failed.

        :raise: Unauthorized Authentication failed.
        :return: The user object.
        """
        raise NotImplementedError()


class TokenAuth(Auth):
    API_DOCS = Auth.API_DOCS.copy()
    API_DOCS['description'] = 'Basic scheme with token.'

    def authenticate(self, token: str, *args, **kw) -> object:
        """
        The result of this method is
        a user or a raised exception if authentication failed.

        :raise: Unauthorized Authentication failed.
        :return The user object.
        """
        raise NotImplementedError()



from setuptools import find_packages, setup

with open('README.md') as f:
    long_description = f.read()

setup(
    name='teal',
    version='0.2.0a18',
    packages=find_packages(),
    url='https://github.com/ereuse/teal',
    license='BSD',
    author='Xavier Bustamante Talavera',
    author_email='xavier@bustawin.com',
    description='RESTful Flask for big applications.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        'anytree',
        'apispec',
        'boltons',
        'ereuse-utils[naming, test]>=0.4.0b6',
        'flask>=1.0',
        'flask-sqlalchemy',
        'sqlalchemy-utils[password, color, phone]',  # todo make it optional
        'marshmallow==3.0.0b11',
        'webargs',
        'flask-cors',
        'click-spinner'
    ],
    tests_requires=[
        'pytest',
        'pytest-datadir'
    ],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP :: HTTP Servers',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
)



from ereuse_utils import JSONEncoder
from flask.json import JSONEncoder as FlaskJSONEncoder
from sqlalchemy.ext.baked import Result
from sqlalchemy.orm import Query


class TealJSONEncoder(JSONEncoder, FlaskJSONEncoder):

    def default(self, obj):
        if isinstance(obj, (Result, Query)):
            return tuple(obj)
        return super().default(obj)



import json
from json import JSONDecodeError

from ereuse_utils import flatten_mixed
from marshmallow import Schema as MarshmallowSchema
from marshmallow.fields import Boolean, Field, List, Nested, Str, missing_
from sqlalchemy import Column, between, or_
from webargs.flaskparser import FlaskParser


class ListQuery(List):
    """Base class for list-based queries."""

    def __init__(self, column: Column, cls_or_instance, **kwargs):
        self.column = column
        super().__init__(cls_or_instance, **kwargs)


class Between(ListQuery):
    """
    Generates a `Between` SQL statement.

    This method wants the user to provide exactly two parameters:
    min and max::

        f = Between(Model.foo, Integer())
        ...
        Query().loads({'f': [0, 100]}

    """

    def _deserialize(self, value, attr, data):
        l = super()._deserialize(value, attr, data)
        return between(self.column, *l)


class Equal(Field):
    """
    Generates an SQL equal ``==`` clause for a given column and value::

        class MyArgs(Query):
            f = Equal(MyModel.foo, Integer())
        MyArgs().load({'f': 24}) -> SQL: ``MyModel.foo == 24``

    """

    def __init__(self, column: Column, field: Field,
                 default=missing_, attribute=None, data_key=None, error=None, validate=None,
                 required=False, allow_none=None, load_only=False, dump_only=False,
                 missing=missing_, error_messages=None, **metadata):
        super().__init__(default, attribute, data_key, error, validate, required, allow_none,
                         load_only, dump_only, missing, error_messages, **metadata)
        self.column = column
        self.field = field

    def _deserialize(self, value, attr, data):
        v = super()._deserialize(value, attr, data)
        return self.column == self.field.deserialize(v)


class Or(List):
    """
    Generates an `OR` SQL statement. This is like a Marshmallow List field,
    so you can specify the type of value of the OR and validations.

    As an example, you can define with this a list of options::

        f = Or(Equal(Model.foo, Str(validates=OneOf(['option1', 'option2'])))

    Where the user can select one or more::

        {'f': ['option1']}

    And with ``Length`` you can enforce the user to only choose one option::

        f = Or(..., validates=Length(equal=1))
    """

    def _deserialize(self, value, attr, data):
        l = super()._deserialize(value, attr, data)
        return or_(v for v in l)


class ILike(Str):
    """
    Generates a insensitive `LIKE` statement for strings.
    """

    def __init__(self, column: Column,
                 default=missing_, attribute=None, data_key=None, error=None, validate=None,
                 required=False, allow_none=None, load_only=False, dump_only=False,
                 missing=missing_, error_messages=None, **metadata):
        super().__init__(default, attribute, data_key, error, validate, required, allow_none,
                         load_only, dump_only, missing, error_messages, **metadata)
        self.column = column

    def _deserialize(self, value, attr, data):
        v = super()._deserialize(value, attr, data)
        return self.column.ilike('{}%'.format(v))


class Join(Nested):
    def __init__(self, join,
                 nested, default=missing_, exclude=tuple(), only=None, **kwargs):
        super().__init__(nested, default, exclude, only, **kwargs)
        self.join = join

    def _deserialize(self, value, attr, data):
        v = list(super()._deserialize(value, attr, data))
        v.append(self.join)
        return v


class Query(MarshmallowSchema):
    """
    A Marshmallow schema that outputs SQLAlchemy queries when ``loading``
    dictionaries::

        class MyQuery(Query):
            foo = Like(Mymodel.foocolumn)

        Mymodel.query.filter(*MyQuery().load({'foo': 'bar'})).all()
        # Executes query SELECT ... WHERE foocolumn IS LIKE 'bar%'

    When used with ``webargs`` library you can pass generate queries
    directly from the browser: ``foo.com/foo/?filter={'foo': 'bar'}``.
    """

    def load(self, data, many=None, partial=None):
        """
        Flatten ``Nested`` ``Query`` and add the list of results to
        a SQL ``AND``.
        """
        values = super().load(data, many, partial).values()
        return flatten_mixed(values)

    def dump(self, obj, many=None, update_fields=True):
        raise NotImplementedError('Why would you want to dump a query?')


class Sort(MarshmallowSchema):
    """
    A Marshmallow schema that outputs SQLAlchemy order clauses::

        class MySort(Sort):
            foo = SortField(MyModel.foocolumn)
        MyModel.query.filter(...).order_by(*MyQuery().load({'foo': 0})).all()

    When used with ``webargs`` library you can pass generate sorts
    directly from the browser: ``foo.com/foo/?sort={'foo': 1, 'bar': 0}``.
    """

    ASCENDING = True
    """Sort in ascending order."""
    DESCENDING = False
    """Sort in descending order."""

    def load(self, data, many=None, partial=None):
        values = super().load(data, many, partial).values()
        return flatten_mixed(values)


class SortField(Boolean):
    """A field that outputs a SQLAlchemy order clause."""

    def __init__(self, column: Column, truthy=Boolean.truthy, falsy=Boolean.falsy, **kwargs):
        super().__init__(truthy, falsy, **kwargs)
        self.column = column

    def _deserialize(self, value, attr, data):
        v = super()._deserialize(value, attr, data)
        return self.column.asc() if v else self.column.desc()


class NestedQueryFlaskParser(FlaskParser):
    """
    Parses JSON-encoded URL parameters like
    ``.../foo?param={"x": "y"}&param2=["x", "y"]``, and it still allows
    normal non-JSON-encoded params ``../foo?param=23&param2={"a": "b"}``.
    """

    def parse_querystring(self, req, name, field):
        v = super().parse_querystring(req, name, field)
        try:
            return json.loads(v)
        except (JSONDecodeError, TypeError):
            return v


class FullTextSearch(Str):
    # todo this is dummy for now
    pass



import ipaddress
from distutils.version import StrictVersion

import pytest
from boltons import urlutils
from flask_sqlalchemy import SQLAlchemy
from werkzeug.exceptions import NotFound

from teal.db import DBError, IP, StrictVersionType, URL, UniqueViolation
from teal.teal import Teal
from tests import conftest


def test_not_found(app: Teal):
    """
    When not finding a resource, the db should raise a ``NotFound``
    exception instead of the built-in for SQLAlchemy.
    """
    with app.app_context():
        Device = app.resources['Device'].MODEL
        with pytest.raises(NotFound):
            Device.query.one()


def test_db_default_column_name(db: SQLAlchemy):
    """Ensures that the default column name is snake case (default)."""

    class Foo(db.Model):
        id = db.Column(db.Integer, primary_key=True)

    assert Foo.__tablename__ == 'foo'

    class FooBar(db.Model):
        id = db.Column(db.Integer, primary_key=True)

    assert FooBar.__tablename__ == 'foo_bar'


def test_db_psql_schemas(db: SQLAlchemy):
    """Tests multiple psql schemas."""
    # todo do this


@pytest.mark.usefixtures(conftest.app_context.__name__)
def test_db_strict_version_type(db: SQLAlchemy):
    class Foo(db.Model):
        id = db.Column(db.Integer, primary_key=True)
        bar = db.Column(StrictVersionType)

    db.create_all()
    foo = Foo(id=1, bar=StrictVersion('1.0.0a1'))
    assert isinstance(foo.bar, StrictVersion)
    db.session.add(foo)
    db.session.commit()


@pytest.mark.usefixtures(conftest.app_context.__name__)
def test_db_url(db: SQLAlchemy):
    class Foo(db.Model):
        id = db.Column(db.Integer, primary_key=True)
        bar = db.Column(URL)

    db.create_all()
    foo = Foo(id=1, bar=urlutils.URL('http://foo.com/bar'))
    assert isinstance(foo.bar, urlutils.URL)
    db.session.add(foo)
    db.session.commit()


def test_db_ip():
    # We cannot actually try this on the db as we use sqlite for testing
    ip = IP()
    x = ip.process_bind_param(ipaddress.ip_address('192.168.1.1'), None)
    assert x == '192.168.1.1'
    y = ip.process_result_value('192.168.1.1', None)
    assert str(y) == '192.168.1.1'


def test_db_error():
    normal = DBError(Exception())
    assert normal.__class__ == DBError
    unique = DBError(Exception('UNIQUE constraint'))
    assert unique.__class__ == UniqueViolation


@pytest.mark.usefixtures(conftest.app_context.__name__)
def test_db_error_unique_violation(db: SQLAlchemy):
    """UniqueViolation"""

    class Foo(db.Model):
        id = db.Column(db.Integer, primary_key=True)
        unique = db.Column(db.Integer, unique=True)

    db.create_all()
    foo = Foo(id=1, unique=1)
    db.session.add(foo)
    db.session.commit()
    foo_again = Foo(id=2, unique=1)
    db.session.add(foo_again)
    with pytest.raises(UniqueViolation):
        db.session.commit()
